#!/usr/bin/env python

# Copyright 2017 Jose Nelson Amaral
# Copyright 2017 Elliot Colp
# Copyright 2017 Erick Ochoa Lopez
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
This script takes random games from the games directory to produce workloads for
the leela_r benchmark.
The executable in this directory is required to run the script, but can be
recompiled for your architecture using the SPEC CPUv6 kit. The leela_r
program can be found in benchspec/CPUv6/541.leela_r/exe.

Usage: ./gen-input.py [options]

This script's output will be placed in ./<workload-name>
"""

import sys, argparse
import os, glob, shutil, subprocess
import re
import random

# Get the available board sizes
sizes = os.walk("./sgf").next()[1]
sizes.sort(key=lambda name: int(name))

# Parse arguments
parser = argparse.ArgumentParser(usage="./gen-input.py [options]",
description="""
Randomly choose games to create workloads for leela_r. See docstring at top of
gen-input.py for more information.
""",
formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument("--name", "-n", help="The workload's name. If not given, the name will be the lowest number from 000 to 999 which doesn't already exist in this directory.",
                    default=argparse.SUPPRESS)
parser.add_argument("--count", "-c", help="The number of games to include.",
                    default=6, type=int)
parser.add_argument("--removes", "-r", help="The minimum and maximum number of moves to take off the end of the game.",
                    nargs=2, default=[90, 120], type=int, metavar=("MIN", "MAX"))
parser.add_argument("--boardsize", "-b", help="The size of the board to use",
                    default=sizes[-1], choices=sizes)
parser.add_argument("--seed", "-s", help="The seed to feed the random number generator before creating the workload. If not given, system time is used.",
                    default=argparse.SUPPRESS)
parser.add_argument("--norun", action="store_true", help="Do not run the workloads once they are created")

parsedArgs, parsedRest = parser.parse_known_args()
    
# Get min/max move removes
minRemoves, maxRemoves = parsedArgs.removes

# Get list of games
gameList = glob.glob("./sgf/" + parsedArgs.boardsize + "/*")

# Get name
if "name" not in parsedArgs:
    wkldName = "000"
    # Find the next number
    for i in range(999):
        iStr = "{:03}".format(i)
        
        if not os.path.isdir(iStr):
            wkldName = iStr
            break
else:
    wkldName = parsedArgs.name

# Read the workload name (stripping newline)
wkldDir = "./" + wkldName
inputDir = wkldDir + "/input"
outputDir = wkldDir + "/output"

print("\nBuilding workload \"" + wkldName + "\"")

# Remove old data
if os.path.isdir(wkldDir):
    resp = raw_input("Directory \"" + wkldDir + "\" already exists. Overwrite? ")
    if resp not in set(["yes", "y"]):
        print("Aborting")
        exit()
    else:
        shutil.rmtree(wkldDir)
    
# Set up directories
os.makedirs(inputDir)
os.makedirs(outputDir)

# Write dummy reftime/refpower files
with open(wkldDir + "/reftime", "w") as f:
    f.write(wkldName + "\n2\n")
    
with open(wkldDir + "/refpower", "w") as f:
    f.write(wkldName + "\n2\n")

# Seed the PRNG
if "seed" in parsedArgs:
    random.seed(parsedArgs.seed)

# Create the input
inputFile = "{}/{}.sgf".format(inputDir, wkldName)
controlFile = "{}/control".format(inputDir)
with open(controlFile, "w") as f:
    f.write("{}.sgf".format(wkldName))

with open(inputFile, "w") as f:
    for i in xrange(parsedArgs.count):
        # Choose a random number of moves to remove
        removeCount = random.randint(minRemoves, maxRemoves)
        
        gameFound = False
        for i in xrange(100):
            # Choose a random game without replacement
            game = random.choice(gameList)
            gameList.remove(game)
            
            # Read the game string
            with open(game, "r") as gameFile:
                gameStr = gameFile.read()
                
            moves = []
            
            # Handle handicap move
            moveStart = gameStr.find("AB[")
            if moveStart > -1:
                # Find white's move
                normalStart = gameStr.find(";W[")
                
                # Include the black handicap placement as a move
                moves.append(gameStr[moveStart:normalStart])
            else:
                # Normal start
                normalStart = gameStr.find(";B[")
                moveStart = normalStart
            
            # Get the list of moves
            moves += (re.findall(r"[BW]\[[^\]]*\]", gameStr[normalStart:]))
            
            # Keep looping if there aren't enough moves
            if len(moves) >= removeCount:
                gameFound = True
                break
        
        # No suitable game was found
        if not gameFound:
            print("After 100 tries, no games were found with at least {} "
                  "moves!".format(removeCount))
                  
        # Write the first line of game info
        lineEnd = gameStr.find("\n") + 1
        f.write(gameStr[:lineEnd])
        
        # Write workload info
        f.write("GC[Original: {} moves; removed {}]\n".format(len(moves),
                                                              removeCount))
        
        # Write up to moves
        f.write(gameStr[lineEnd:moveStart])
        
        # Write the moves (max 10 per line)
        moveCount = len(moves) - removeCount
        
        lineCount = 10
        for n in range(0, moveCount, lineCount):
            f.write(";")
            f.write(";".join(moves[n:n + lineCount]))
            f.write("\n")

        f.write(";)\n\n")

